/*
 *  diskdumpfmt.c
 *
 *  Copyright (C) 2004  FUJITSU LIMITED
 *  Written by Akira Imamura (imamura.akira@jp.fujitsu.com)
 */
/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */
/* $Id: diskdump_format.c,v 1.25 2004/10/13 18:50:40 tachino Exp $ */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <linux/fs.h>
#include <stdint.h>
#include <sys/time.h>
#include <signal.h>
#include <time.h>
#include <assert.h>
#include "diskdump_format.h"

#define DUMP_PARTITION_SIGNATURE "diskdump"
#define Formula(x) ((x + 7) * 11)

static int set_block_size(dump_device_t *device)
{
	struct stat stt;
	off_t block_size;

	assert(device->page_size);

	if (fstat(device->fd, &stt) < 0) {
		perror("fstat");
		return -1;
	}

	if (S_ISBLK(stt.st_mode) || S_ISCHR(stt.st_mode)) {
		unsigned long sz;

		if (ioctl(device->fd, BLKGETSIZE, &sz) < 0) {
			perror("ioctl(BLKGETSIZE)");
			return -1;
		}

		block_size = sz / (device->page_size / 512);
	} else {
		block_size = stt.st_size / device->page_size;
	}

	device->block_size = block_size;

	return 0;
}

extern int posix_memalign(void **, size_t, size_t );

dump_device_t *open_device(const char *dev_name)
{
	int fd;
	dump_device_t *device;
	int page_size;
	void *buf;

	device = calloc(1, sizeof(*device));
	if (device == NULL)
		return NULL;

	if ((fd = open(dev_name, O_RDWR)) == -1) {
		free(device);
		return NULL;
	}
	device->fd = fd;

	if ((page_size = sysconf(_SC_PAGE_SIZE)) < 0)
		goto bad;

	device->page_size = page_size;

	if (posix_memalign(&buf, page_size, page_size) != 0)
		goto bad;
	device->scratch = buf;

	if (set_block_size(device) < 0)
		goto bad;

	return device;
bad:
	delete_device(device);
	return NULL;
}

void delete_device(dump_device_t *device)
{
	if (device->fd < 0)
		close(device->fd);
	free(device->scratch);
	free(device);
}

int get_page_size(dump_device_t *device)
{
	return device->page_size;
}

off_t get_block_size(dump_device_t *device)
{
	return device->block_size;
}

void set_progress(dump_device_t *device, int do_progress)
{
	device->do_progress = do_progress;
}

/*
 * 1st word = 'disk'
 * 2nd word = 'dump'
 * 3rd word = block number [0...n]
 * 4th word = (block number * 11)
 * 5th word = (4th word * 11)
 * ..
 *
 * The device is checked out from the head of that.
 */

int check_diskdump_signature(dump_device_t *device, int page_no)
{
	int word_cnt = sizeof(DUMP_PARTITION_SIGNATURE) - 1;
	int i;
	uint32_t *buf = device->scratch;
	uint32_t val = page_no;
	int page_size = device->page_size;

	if (lseek(device->fd, (off_t)page_no * page_size, SEEK_SET) == -1) {
		perror("lseek");
		return -1;
	}
	if (read(device->fd, buf, page_size) < page_size)
		return 255;

	if (memcmp(buf, DUMP_PARTITION_SIGNATURE, word_cnt) != 0)
		return -1;

	for (i = 3; i < page_size / sizeof(int); i++) {
		val = Formula(val);
		if (buf[i] != val)
			return -1;
	}

	return 0;
}

struct progress_bar {
	int	blk_ini;
	int	blk_end;
	time_t	last_time;
};

/*
 * 1st word = 'disk'
 * 2nd word = 'dump'
 * 3rd word = block number [0...n]
 * 4th word = (block number * 11)
 * 5th word = (4th word * 11)
 * ..
 *
 * The device is formatted according to the above rule.
 */

static void print_progress(struct progress_bar *pbar, int blk_nr)
{
	double progress;
	time_t tm;

	if (pbar == NULL)
		return;

	tm = time(NULL);

	if (tm - pbar->last_time < 1)
		return;

	pbar->last_time = tm;

	progress = (double)blk_nr * 100.0
		 / (double)pbar->blk_end;

	fprintf(stderr, "progress     : %5.1f%%\r", progress);
}

static void print_progress2(struct progress_bar *pbar, int blk_nr)
{
	double progress;

	if (pbar == NULL)
		return;

	progress = (double)blk_nr * 100.0
		 / (double)pbar->blk_end;

	fprintf(stderr, "progress     : %5.1f%%\r", progress);

	if (pbar->blk_end == blk_nr)
		fprintf(stderr, "\n");
}

static int format(dump_device_t *device, int blk_ini, int blk_end)
{
	uint32_t *words = device->scratch;
	int blk_nr;
	int val;
	int i;
	struct progress_bar *pbar = NULL;
	int page_size = device->page_size;
	
	if (device->do_progress && (blk_end - blk_ini > 1))
		pbar = &(struct progress_bar){
			.blk_ini	= blk_ini,
			.blk_end	= blk_end,
			.last_time	= time(NULL),
		};

	if (lseek(device->fd, page_size * blk_ini, SEEK_SET) == -1) {
		perror("lseek");
		return -1;
	}

	print_progress(pbar, 0);

	for (blk_nr = blk_ini; blk_nr < blk_end; blk_nr++) {
		memcpy(words, DUMP_PARTITION_SIGNATURE,
				sizeof(DUMP_PARTITION_SIGNATURE) - 1);

		words[2] = blk_nr;

		val = Formula(blk_nr);
		for (i = 3; i < page_size / sizeof(words[0]); i++) {
			words[i] = val;
			val = Formula(val);
		}
		if (write(device->fd, words, page_size) == -1)
			return -1;

		print_progress(pbar, blk_nr);
	}
	print_progress2(pbar, blk_nr);

	sync();

	return 0;
}

/*
 * In order not to make the device which was formatted incompletely even if
 * diskdumpfmt aborted,
 * the device's header is formatted last.
 */

int write_diskdump_signature(dump_device_t *device)
{
	if (device->block_size <= 5)
		return -1;

	if (format(device, 2, device->block_size) == -1)
		return -1;
	if (format(device, 1, 2) == -1)
		return -1;
	if (format(device, 0, 1) == -1)
		return -1;

	return 0;
}

